#! /usr/bin/env python
import platform
import re

from downward import suites
from downward.reports.absolute import AbsoluteReport
from lab.environments import TetralithEnvironment, BaselSlurmEnvironment, LocalEnvironment
from lab.experiment import Experiment
from lab.reports import Attribute, geometric_mean

# Create custom report class with suitable info and error attributes.
class BaseReport(AbsoluteReport):
    INFO_ATTRIBUTES = ["time_limit", "memory_limit"]
    ERROR_ATTRIBUTES = [
        "domain",
        "problem",
        "algorithm",
        "unexplained_errors",
        "error",
        "node",
    ]

NODE = platform.node()
print(NODE)
REMOTE = re.match(r"tetralith\d+.nsc.liu.se|n\d+", NODE)

BENCHMARKS_DIR = "../benchmarks"
print(REMOTE)
if REMOTE:
    ENV = TetralithEnvironment(
        partition="tetralith",
        email="",
        memory_per_cpu="3G",
        cpus_per_task=16,
        setup=TetralithEnvironment.DEFAULT_SETUP,
        extra_options="#SBATCH --account=snic2022-5-341")
    SUITE = ["barman", "blocksworld_3", "blocksworld_4", "childsnack", "delivery", "gripper", "miconic", "reward", "spanner", "visitall"]
    TIME_LIMIT = 3 * 3600
else:
    ENV = LocalEnvironment(processes=4)
    SUITE = ["blocksworld_3:p-3-0.pddl", "childsnack:p-2-1.0-0.0-1-0.pddl", "delivery:instance_2_1_0.pddl", "gripper:p-1-0.pddl", "miconic:p-2-2-0.pddl", "reward:instance_2x2_0.pddl", "visitall:p-1-0.5-2-0.pddl"]
    TIME_LIMIT = 180
ATTRIBUTES = [
    Attribute("num_iterations", absolute=True, min_wins=True, scale="linear"),
    Attribute("num_states", absolute=True, min_wins=True, scale="linear"),
    Attribute("num_dynamic_atoms", absolute=True, min_wins=True, scale="linear"),
    Attribute("num_static_atoms", absolute=True, min_wins=True, scale="linear"),
    Attribute("generate_time_complexity_5", absolute=True, min_wins=True, scale="linear"),
    Attribute("generate_memory_complexity_5", absolute=True, min_wins=True, scale="linear"),
    Attribute("generate_time_complexity_10", absolute=True, min_wins=True, scale="linear"),
    Attribute("generate_memory_complexity_10", absolute=True, min_wins=True, scale="linear"),
    Attribute("num_booleans", absolute=True, min_wins=True, scale="linear"),
    Attribute("num_numericals", absolute=True, min_wins=True, scale="linear"),
    Attribute("evaluate_time_single", absolute=True, min_wins=True, scale="linear"),
    Attribute("evaluate_time_single_cache", absolute=True, min_wins=True, scale="linear"),
    Attribute("evaluate_time_multi_cache", absolute=True, min_wins=True, scale="linear"),

]

MEMORY_LIMIT = (16 * 3000) * 0.98

GENERATOR_TIME_LIMIT = 2 * 3600
GENERATOR_FEATURE_LIMIT = 1000000
NUM_FEATURE_VALUATION_ITERATIONS = 1000

# Create a new experiment.
exp = Experiment(environment=ENV)
# Add custom parser for FF.
exp.add_parser("experiment_parser.py")

for task in suites.build_suite(BENCHMARKS_DIR, SUITE):
    for complexity in [10]:
        run = exp.add_run()
        # Create symbolic links and aliases. This is optional. We
        # could also use absolute paths in add_command().
        run.add_resource("domain", task.domain_file, symlink=True)
        run.add_resource("problem", task.problem_file, symlink=True)
        run.add_resource("experiment_generator", "experiment_generator", symlink=True)
        # 'ff' binary has to be on the PATH.
        # We could also use exp.add_resource().
        run.add_command(
            f"complexity-{complexity}",
            ["./experiment_generator", "{domain}", "{problem}", complexity, complexity, complexity, complexity, complexity, GENERATOR_TIME_LIMIT, GENERATOR_FEATURE_LIMIT, NUM_FEATURE_VALUATION_ITERATIONS],
            time_limit=TIME_LIMIT,
            memory_limit=MEMORY_LIMIT,
        )
        # AbsoluteReport needs the following properties:
        # 'domain', 'problem', 'algorithm', 'coverage'.
        run.set_property("domain", task.domain)
        run.set_property("problem", task.problem)
        run.set_property("algorithm", f"complexity-{complexity}")
        # BaseReport needs the following properties:
        # 'time_limit', 'memory_limit'.
        run.set_property("time_limit", TIME_LIMIT)
        run.set_property("memory_limit", MEMORY_LIMIT)
        # Every run has to have a unique id in the form of a list.
        # The algorithm name is only really needed when there are
        # multiple algorithms.
        run.set_property("id", [f"complexity-{complexity}", task.domain, task.problem])

# Add step that writes experiment files to disk.
exp.add_step("build", exp.build)

# Add step that executes all runs.
exp.add_step("start", exp.start_runs)

# Add step that collects properties from run directories and
# writes them to *-eval/properties.
exp.add_fetcher(name="fetch")

# Make a report.
exp.add_report(BaseReport(attributes=ATTRIBUTES), outfile="report.html")

# Parse the commandline and run the specified steps.
exp.run_steps()


from typing import Overload, List, Tuple


class ConceptDenotation:
    def __init__(self, num_objects: int) -> None: ...
    def __eq__(self) -> bool: ...
    def __ne__(self) -> bool: ...
    def __str__(self) -> str: ...
    def __len__(self) -> int: ...
    def __hash__(self) -> int: ...
    def contains(self, object_index: int)-> bool: ...
    def set(self) -> None: ...
    def insert(self, object_index: int) -> None: ...
    def erase(self, object_index: int) -> None: ...
    def intersects(self, other: "ConceptDenotation") -> bool: ...
    def is_subset_of(self, other: "ConceptDenotation") -> bool: ...
    def to_vector(self) -> List[int]: ...
    def to_sorted_vector(self) -> List[int]: ...
    def get_num_objects(self) -> int: ...


class RoleDenotation:
    def __init__(self, num_objects: int) -> None: ...
    def __eq__(self) -> bool: ...
    def __ne__(self) -> bool: ...
    def __str__(self) -> str: ...
    def __len__(self) -> int: ...
    def __hash__(self) -> int: ...
    def contains(self, pair_of_object_indices: Tuple[int, int])-> bool: ...
    def set(self) -> None: ...
    def insert(self, pair_of_object_indices: Tuple[int, int]) -> None: ...
    def erase(self, pair_of_object_indices: Tuple[int, int]) -> None: ...
    def intersects(self, other: "ConceptDenotation") -> bool: ...
    def is_subset_of(self, other: "ConceptDenotation") -> bool: ...
    def to_vector(self) -> List[Tuple[int, int]]: ...
    def to_sorted_vector(self) -> List[Tuple[int, int]]: ...
    def get_num_objects(self) -> int: ...


class DenotationsCaches:
    def __init__(self) -> None: ...


class Constant:
    def __eq__(self, other: "Constant") -> bool: ...
    def __ne__(self, other: "Constant") -> bool: ...
    def __str__(self) -> str: ...
    def get_index(self) -> int: ...
    def get_name(self) -> str: ...


class Predicate:
    def __eq__(self, other: "Predicate") -> bool: ...
    def __ne__(self, other: "Predicate") -> bool: ...
    def __str__(self) -> str: ...
    def get_index(self) -> int: ...
    def get_name(self) -> str: ...
    def get_arity(self) -> int: ...
    def is_static(self) -> bool: ...


class VocabularyInfo:
    def __init__(self) -> None: ...
    def __str__(self) -> str: ...
    def add_predicate(self, name: str, arity: int, is_static: bool) -> Predicate: ...
    def add_constant(self, name: str) -> Constant: ...
    def get_predicates(self) -> List[Predicate]: ...
    def get_constants(self) -> List[Constant]: ...
    def get_predicate(self, name: str) -> Predicate: ...
    def get_constant(self, name: str) -> Constant: ...


class Object:
    def __eq__(self, other: "Object") -> bool: ...
    def __ne__(self, other: "Object") -> bool: ...
    def __str__(self) -> str: ...
    def get_index(self) -> int: ...
    def get_name(self) -> str: ...


class Atom:
    def __eq__(self, other: "Atom") -> bool: ...
    def __ne__(self, other: "Atom") -> bool: ...
    def __str__(self) -> str: ...
    def get_index(self) -> int: ...
    def get_name(self) -> str: ...
    def get_predicate_index(self) -> int: ...
    def get_object_indices(self) -> List[int]: ...
    def is_static(self) -> bool: ...


class InstanceInfo:
    def __init__(self, index: int, vocabulary: VocabularyInfo) -> None: ...
    def __str__(self) -> str: ...
    def add_object(self, name: str) -> Object: ...
    @overload
    def add_atom(self, predicate: Predicate, objects: List[Object]) -> Atom: ...
    @overload
    def add_atom(self, predicate_name: str, object_names: List[str]) -> Atom: ...
    @overload
    def add_atom(self, predicate_index: int, object_indices: List[int]) -> Atom: ...
    @overload
    def add_static_atom(self, predicate: Predicate, objects: List[Object]) -> Atom: ...
    @overload
    def add_static_atom(self, predicate_name: str, object_names: List[str]) -> Atom: ...
    @overload
    def add_static_atom(self, predicate_index: int, object_indices: List[int]) -> Atom: ...
    def get_index(self) -> int: ...
    def get_objects(self) -> List[Object]: ...
    def get_atoms(self) -> List[Atom]: ...
    def get_static_atoms(self) -> List[Atom]: ...
    def get_object(self, object_name: str) -> Object: ...
    def get_atom(self, atom_name: str) -> Atom: ...
    def get_vocabulary_info(self) -> VocabularyInfo: ...
    def clear_atoms(self) -> None: ...
    def clear_static_atoms(self) -> None: ...


class State:
    @overload
    def __init__(self, instance: InstanceInfo, atoms: List[Atom]) -> None: ...
    @overload
    def __init__(self, instance: InstanceInfo, atom_indices: List[int]) -> None: ...
    def __eq__(self, other: "State") -> bool: ...
    def __ne__(self, other: "State") -> bool: ...
    def __str__(self) -> str: ...
    def __hash__(self) -> int: ...
    def get_index(self) -> int: ...
    def get_atom_indices(self) -> List[int]: ...
    def get_instance_info(self) -> InstanceInfo: ...


class Concept():
    def __str__(self) -> str: ...
    def compute_complexity(self) -> int: ...
    def get_index(self) -> int: ...
    def get_vocabulary_info(self) -> VocabularyInfo: ...
    @overload
    def evaluate(self, state: State) -> ConceptDenotation: ...
    @overload
    def evaluate(self, state: State, denotations_caches: DenotationsCaches) -> ConceptDenotation: ...
    @overload
    def evaluate(self, states: List[State], denotations_caches: DenotationsCaches) -> List[ConceptDenotation]: ...


class Role():
    def __str__(self) -> str: ...
    def compute_complexity(self) -> int: ...
    def get_index(self) -> int: ...
    def get_vocabulary_info(self) -> VocabularyInfo: ...
    @overload
    def evaluate(self, state: State) -> RoleDenotation: ...
    @overload
    def evaluate(self, state: State, denotations_caches: DenotationsCaches) -> RoleDenotation: ...
    @overload
    def evaluate(self, states: List[State], denotations_caches: DenotationsCaches) -> List[RoleDenotation]: ...


class Boolean():
    def __str__(self) -> str: ...
    def compute_complexity(self) -> int: ...
    def get_index(self) -> int: ...
    def get_vocabulary_info(self) -> VocabularyInfo: ...
    @overload
    def evaluate(self, state: State) -> bool: ...
    @overload
    def evaluate(self, state: State, denotations_caches: DenotationsCaches) -> bool: ...
    @overload
    def evaluate(self, states: List[State], denotations_caches: DenotationsCaches) -> List[bool]: ...


class Numerical():
    def __str__(self) -> str: ...
    def compute_complexity(self) -> int: ...
    def get_index(self) -> int: ...
    def get_vocabulary_info(self) -> VocabularyInfo: ...
    @overload
    def evaluate(self, state: State) -> int: ...
    @overload
    def evaluate(self, state: State, denotations_caches: DenotationsCaches) -> int: ...
    @overload
    def evaluate(self, states: List[State], denotations_caches: DenotationsCaches) -> List[int]: ...


class SyntacticElementFactory:
    def __init__(self, vocabulary: VocabularyInfo) -> None: ...
    def parse_concept(self, description: str, filename: str = "") -> Concept: ...
    def parse_role(self, description: str, filename: str = "") -> Role: ...
    def parse_boolean(self, description: str, filename: str = "") -> Boolean: ...
    def parse_numerical(self, description: str, filename: str = "") -> Numerical: ...

    @overload
    def make_empty_boolean(self, concept: Concept) -> Boolean: ...
    @overload
    def make_empty_boolean(self, role: Role) -> Boolean: ...
    @overload
    def make_inclusion_boolean(self, concept_left: Concept, concept_right: Concept) -> Boolean: ...
    @overload
    def make_inclusion_boolean(self, role_left: Role, role_right: Concept) -> Boolean: ...
    def make_nullary_boolean(self, predicate: Predicate) -> Boolean: ...

    def make_all_concept(self, role: Role, concept: Concept) -> Concept: ...
    def make_and_concept(self, concept_left: Concept, concept_right: Concept) -> Concept: ...
    def make_bot_concept(self) -> Concept: ...
    def make_diff_concept(self, concept_left: Concept, concept_right: Concept) -> Concept: ...
    def make_equal_concept(self, role_left: Role, role_right: Role) -> Concept: ...
    def make_not_concept(self, concept: Concept) -> Concept: ...
    def make_one_of_concept(self, constant: Constant) -> Concept: ...
    def make_or_concept(self, concept_left: Concept, concept_right: Concept) -> Concept: ...
    def make_projection_concept(self, role: Role, position: int) -> Concept: ...
    def make_primitive_concept(self, predicate: Predicate, position: int) -> Concept: ...
    def make_some_concept(self, role: Role, concept: Concept) -> Concept: ...
    def make_subset_concept(self, role_left: Role, role_right: Role) -> Concept: ...
    def make_top_concept(self) -> Concept: ...

    def make_concept_distance_numerical(self, concept_from: Concept, role: Role, concept_to: Concept) -> Numerical: ...
    @overload
    def make_count_numerical(self, concept: Concept) -> Numerical: ...
    @overload
    def make_count_numerical(self, role: Role) -> Numerical: ...
    def make_role_distance_numerical(self, role_from: Role, role: Role, role_to: Role) -> Numerical: ...
    def make_sum_concept_distance_numerical(self, concept_from: Concept, role: Role, concept_to: Concept) -> Numerical: ...
    def make_sum_role_distance_numerical(self, role_from: Role, role: Role, role_to: Role) -> Numerical: ...

    def make_and_concept(self, role_left: Role, role_right: Role) -> Role: ...
    def make_compose_role(self, role_left: Role, role_right: Role) -> Role: ...
    def make_diff_role(self, role_left: Role, role_right: Role) -> Role: ...
    def make_identity_role(self, concept: Concept) -> Role: ...
    def make_inverse_role(self, role: Role) -> Role: ...
    def make_not_role(self, role: Role) -> Role: ...
    def make_or_concept(self, role_left: Role, role_right: Role) -> Role: ...
    def make_primitive_role(self, predicate: Predicate, position_first: int, position_second: int) -> Role: ...
    def make_restrict_role(self, role: Role, concept: Concept) -> Role: ...
    def make_top_role(self) -> Role: ...
    def make_transitive_closure(self, role: Role) -> Role: ...
    def make_transitive_reflexive_closure(self, role: Role) -> Role: ...

    def get_vocabulary_info(self): VocabularyInfo: ...


#!/usr/bin/python3

"""
Example illustrating the core component.
"""

from dlplan.core import VocabularyInfo, InstanceInfo, State, \
    SyntacticElementFactory, DenotationsCaches


def construct_vocabulary_info():
    """ Construct a VocabularyInfo for the Blocks domain.

    Create an empty VocabularyInfo and then add predicates, and constants.

    Returns: the VocabularyInfo
    """
    vocabulary = VocabularyInfo()
    vocabulary.add_predicate("on", 2)
    vocabulary.add_predicate("on_g", 2)
    vocabulary.add_predicate("ontable", 1)
    vocabulary.add_predicate("holding", 1)
    vocabulary.add_predicate("clear", 1)
    vocabulary.add_predicate("arm-empty", 0)
    return vocabulary


def construct_instance_info(vocabulary):
    """ Construct an InstanceInfo over the Blocks domain.

    Create an empty InstanceInfo and then add objects, atoms, and static atoms.

    Returns: the InstanceInfo
    """
    # User must ensure that each InstanceInfo gets its unique index for caching.
    instance = InstanceInfo(0, vocabulary)
    instance.add_atom("on", ["a", "b"])
    instance.add_atom("on", ["b", "a"])
    instance.add_atom("ontable", ["a"])
    instance.add_atom("ontable", ["b"])
    instance.add_atom("holding", ["a"])
    instance.add_atom("holding", ["b"])
    instance.add_atom("clear", ["a"])
    instance.add_atom("clear", ["b"])
    instance.add_atom("arm-empty", [])
    instance.add_static_atom("on_g", ["a", "b"])
    return instance


def main():
    """ Example illustrating the core component on a fragment of a planning
    problem over the Blocks domain.

    The example illustrates the creation of VocabularyInfo, InstanceInfo,
    State, Element, and the evaluation of Element on a State using
    DenotationsCaches.
    """
    vocabulary = construct_vocabulary_info()
    instance = construct_instance_info(vocabulary)
    factory = SyntacticElementFactory(vocabulary)

    atoms = instance.get_atoms()
    atom_0 = atoms[0]
    atom_1 = atoms[1]
    atom_3 = atoms[3]
    atom_6 = atoms[6]
    # User must ensure that each State gets its unique index for caching.
    state_0 = State(0, instance, [atom_0, atom_3, atom_6])  # a on b
    state_1 = State(1, instance, [atom_1, atom_3, atom_6])  # b on a
    states = [state_0, state_1]

    numerical = factory.parse_numerical("n_count(c_and(c_primitive(on_g,0),c_primitive(on,0)))")
    print("Element repr:", repr(numerical))
    print()

    print("Evaluate for single state without cache.")
    print("State:", str(state_0))
    print("Value:", numerical.evaluate(state_0))
    print()

    caches = DenotationsCaches()
    print("Evaluate for single state with cache.")
    print("State:", str(state_1))
    print("Value:", numerical.evaluate(state_1, caches))
    print()

    print("Evaluate for multiple states with cache.")
    denotations = numerical.evaluate(states, caches)
    for state, value in zip(states, denotations):
        print("State:", str(state))
        print("Value:", value)


if __name__ == "__main__":
    main()



#!/usr/bin/python3

"""
Example illustrating the novelty component.
"""

from dlplan.novelty import NoveltyBase, NoveltyTable


def novelty_table_example():
    """
    The example illustrates the creation and usage of NoveltyBase and NoveltyTable.
    """
    novelty_base = NoveltyBase(4, 2)
    novelty_table = NoveltyTable(novelty_base)

    # 1. Exhaustively insert all tuple indices
    has_novel = novelty_table.insert_atom_indices([0,1], stop_if_novel=False)
    assert has_novel
    # All tuple indices are not novel anymore
    has_novel = novelty_table.insert_atom_indices([0,1], stop_if_novel=False)
    assert not has_novel

    # 2. Lazy insertion of tuple indices
    has_novel = novelty_table.insert_atom_indices([1,2,3], stop_if_novel=True)
    assert has_novel
    # All tuple indices are not novel anymore
    has_novel = novelty_table.insert_atom_indices([1,2,3], stop_if_novel=True)
    assert has_novel

    # 3. Exhaustively insert all tuple indices with consideration of add atom indices.
    has_novel = novelty_table.insert_atom_indices([0,2], [1,3], stop_if_novel=False)
    assert has_novel


if __name__ == "__main__":
    novelty_table_example()



from typing import List, Overload

from ..state_space import StateSpace


class NoveltyBase:
    def __init__(self, num_atoms: int, arity: int) -> None: ...
    def atom_indices_to_tuple_index(self, atom_indices: List[int]) -> int: ...
    def tuple_index_to_atom_indices(self, tuple_index: int) -> List[int]: ...
    def get_num_atoms(self) -> int: ...
    def get_arity(self) -> int: ...


class NoveltyTable:
    def __init__(self, novelty_base: NoveltyBase) -> None: ...
    @overload
    def compute_novel_tuple_indices(self, atom_indices: List[int]) -> List[int]: ...
    @overload
    def compute_novel_tuple_indices(self, atom_indices: List[int], add_atom_indices: List[int]) -> List[int]: ...
    @overload
    def insert_atom_indices(self, atom_indices: List[int], stop_if_novel: bool = False) -> bool: ...
    @overload
    def insert_atom_indices(self, atom_indices: List[int], add_atom_indices: List[int], stop_if_novel: bool = False) -> bool: ...
    @overload
    def insert_tuple_indices(self, tuple_indices: List[int], stop_if_novel: bool = False) -> bool: ...
    def resize(self, novelty_base: NoveltyBase) -> None: ...
    def get_novelty_base(self) -> NoveltyBase: ...


class TupleNode:
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def get_index(self) -> int: ...
    def get_tuple_index(self) -> int: ...
    def get_state_indices(self) -> List[int]: ...
    def get_predecessors(self) -> List[int]: ...
    def get_successors(self) -> List[int]: ...


class TupleGraph:
    def __init__(self, novelty_base: NoveltyBase, state_space: StateSpace, root_state_index: int) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def to_dot(self, verbosity_level: int) -> str: ...
    def get_novelty_base(self) -> NoveltyBase: ...
    def get_state_space(self) -> StateSpace: ...
    def get_root_state_index(self) -> int: ...
    def get_tuple_nodes(self) -> List[TupleNode]: ...
    def get_tuple_node_indices_by_distance(self) -> List[List[int]]: ...
    def get_state_indices_by_distance(self) -> List[List[int]]: ...



#!/usr/bin/python3

"""
Example illustrating the generator component.
"""

from dlplan.core import VocabularyInfo, InstanceInfo, State, SyntacticElementFactory
from dlplan.generator import generate_features


def construct_vocabulary_info():
    """ Construct a VocabularyInfo for the Blocks domain.

    Create an empty VocabularyInfo and then add predicates, and constants.

    Returns: the VocabularyInfo
    """
    vocabulary = VocabularyInfo()
    vocabulary.add_predicate("on", 2)
    vocabulary.add_predicate("on_g", 2)
    vocabulary.add_predicate("ontable", 1)
    vocabulary.add_predicate("holding", 1)
    vocabulary.add_predicate("clear", 1)
    vocabulary.add_predicate("arm-empty", 0)
    return vocabulary


def construct_instance_info(vocabulary):
    """ Construct an InstanceInfo over the Blocks domain.

    Create an empty InstanceInfo and then add objects, atoms, and static atoms.

    Returns: the InstanceInfo
    """
    # User must ensure that each index gets its unique index for caching.
    instance = InstanceInfo(0, vocabulary)
    instance.add_atom("on", ["a", "b"])
    instance.add_atom("on", ["b", "a"])
    instance.add_atom("ontable", ["a"])
    instance.add_atom("ontable", ["b"])
    instance.add_atom("holding", ["a"])
    instance.add_atom("holding", ["b"])
    instance.add_atom("clear", ["a"])
    instance.add_atom("clear", ["b"])
    instance.add_atom("arm-empty", [])
    instance.add_static_atom("on_g", ["a", "b"])
    return instance


def main():
    """ Example illustrating the generator component on a fragment of a planning
    problem over the Blocks domain.
    """
    vocabulary = construct_vocabulary_info()
    instance = construct_instance_info(vocabulary)
    factory = SyntacticElementFactory(vocabulary)

    atoms = instance.get_atoms()
    atom_0 = atoms[0]
    atom_1 = atoms[1]
    atom_2 = atoms[2]
    atom_3 = atoms[3]
    atom_4 = atoms[4]
    atom_5 = atoms[5]
    atom_6 = atoms[6]
    atom_7 = atoms[7]
    atom_8 = atoms[8]
    # User must ensure that each State gets its unique index for caching.
    state_0 = State(0, instance, [atom_0, atom_3, atom_6, atom_8])  # a on b
    state_1 = State(1, instance, [atom_1, atom_2, atom_7, atom_8])  # b on a
    state_2 = State(2, instance, [atom_2, atom_3, atom_6, atom_7, atom_8])  # a,b on table
    state_3 = State(3, instance, [atom_3, atom_4, atom_7])  # holding a, b on table
    state_4 = State(4, instance, [atom_2, atom_5, atom_6])  # holding b, a on table
    states = [state_0, state_1, state_2, state_3, state_4]

    features = generate_features(factory, states, 5, 5, 10, 10, 10, 180, 100000)

    for feature in features:
        print(feature)


if __name__ == "__main__":
    main()



from typing import List, Overload, Union, MutableSet, Tuple

from ..core import State, DenotationsCaches, Boolean, Numerical, Concept, Role, SyntacticElementFactory



class NamedBoolean():
    def __str__(self) -> str: ...
    def get_key(self) -> str: ...
    def get_element(self) -> Boolean: ...


class NamedNumerical():
    def __str__(self) -> str: ...
    def get_key(self) -> str: ...
    def get_element(self) -> Numerical: ...


class NamedConcept():
    def __str__(self) -> str: ...
    def get_key(self) -> str: ...
    def get_element(self) -> Concept: ...


class NamedRole():
    def __str__(self) -> str: ...
    def get_key(self) -> str: ...
    def get_element(self) -> Role: ...


class BaseCondition:
    def __str__(self) -> str: ...
    @overload
    def evaluate(self, state: State) -> bool: ...
    @overload
    def evaluate(self, state: State, caches: DenotationsCaches) -> bool: ...
    def get_index(self) -> int: ...


class PositiveBooleanCondition(BaseCondition):
    def get_named_element(self) -> NamedBoolean: ...


class NegativeBooleanCondition(BaseCondition):
    def get_named_element(self) -> NamedBoolean: ...


class GreaterNumericalCondition(BaseCondition):
    def get_named_element(self) -> NamedNumerical: ...


class EqualNumericalCondition(BaseCondition):
    def get_named_element(self) -> NamedNumerical: ...


class BaseEffect:
    def __str__(self) -> str: ...
    @overload
    def evaluate(self, source_state: State, target_state: State) -> bool: ...
    @overload
    def evaluate(self, source_state: State, target_state: State, caches: DenotationsCaches) -> bool: ...
    def get_index(self) -> int: ...


class PositiveBooleanEffect(BaseEffect):
    def get_named_element(self) -> NamedBoolean: ...


class NegativeBooleanEffect(BaseEffect):
    def get_named_element(self) -> NamedBoolean: ...


class UnchangedBooleanEffect(BaseEffect):
    def get_named_element(self) -> NamedBoolean: ...


class DecrementNumericalEffect(BaseEffect):
    def get_named_element(self) -> NamedNumerical: ...


class IncrementNumericalEffect(BaseEffect):
    def get_named_element(self) -> NamedNumerical: ...


class UnchangedNumericalEffect(BaseEffect):
    def get_named_element(self) -> NamedNumerical: ...


class Rule:
    def __str__(self) -> str: ...
    @overload
    def evaluate_conditions(self, state: State) -> bool: ...
    @overload
    def evaluate_conditions(self, state: State, caches: DenotationsCaches) -> bool: ...
    @overload
    def evaluate_effects(self, source_state: State, target_state: State) -> bool: ...
    @overload
    def evaluate_effects(self, source_state: State, target_state: State, caches: DenotationsCaches) -> bool: ...
    def get_index(self) -> int: ...
    def get_conditions(self) -> MutableSet[BaseCondition]: ...
    def get_effects(self) -> MutableSet[BaseEffect]: ...


class Policy:
    def __str__(self) -> str: ...
    @overload
    def evaluate(self, source_state: State, target_state: State) -> Union[None, Rule]: ...
    @overload
    def evaluate(self, source_state: State, target_state: State, caches: DenotationsCaches) -> Union[None, Rule]: ...
    @overload
    def evaluate_conditions(self, state: State) -> List[Rule]: ...
    @overload
    def evaluate_conditions(self, state: State, caches: DenotationsCaches) -> List[Rule]: ...
    @overload
    def evaluate_effects(self, source_state: State, target_state: State, rules: List[Rule]) -> bool: ...
    @overload
    def evaluate_effects(self, source_state: State, target_state: State, rules: List[Rule], caches: DenotationsCaches) -> bool: ...
    def get_rules(self) -> MutableSet[Rule]: ...
    def get_booleans(self) -> MutableSet[NamedBoolean]: ...
    def get_numericals(self) -> MutableSet[NamedNumerical]: ...
    def get_concepts(self) -> MutableSet[NamedConcept]: ...


class PolicyFactory:
    def __init__(self, element_factory: SyntacticElementFactory) -> None: ...
    def make_boolean(self, key: str, boolean: Boolean) -> NamedBoolean: ...
    def make_numerical(self, key: str, numerical: Numerical) -> NamedNumerical: ...
    def make_concept(self, key: str, concept: Concept) -> NamedConcept: ...
    def make_role(self, key: str, role: Role) -> NamedRole: ...
    def make_pos_condition(self, boolean: NamedBoolean) -> BaseCondition: ...
    def make_neg_condition(self, boolean: NamedBoolean) -> BaseCondition: ...
    def make_gt_condition(self, numerical: NamedNumerical) -> BaseCondition: ...
    def make_eq_condition(self, numerical: NamedNumerical) -> BaseCondition: ...
    def make_pos_effect(self, boolean: NamedBoolean) -> BaseEffect: ...
    def make_neg_effect(self, boolean: NamedBoolean) -> BaseEffect: ...
    def make_inc_effect(self, bnumerical: NamedNumerical) -> BaseEffect: ...
    def make_dec_effect(self, numerical: NamedNumerical) -> BaseEffect: ...
    @overload
    def make_bot_effect(self, boolean: NamedBoolean) -> BaseEffect: ...
    @overload
    def make_bot_effect(self, numerical: NamedNumerical) -> BaseEffect: ...
    def make_rule(self, conditions: MutableSet[BaseCondition], effects: MutableSet[BaseEffect]) -> Rule: ...
    def make_policy(self, rules: MutableSet[Rule]) -> Policy: ...
    def parse_policy(self, description: str, filename: str = "") -> Policy: ...


class PolicyMinimizer:
    def __init__(self) -> None: ...
    @overload
    def minimize(self, policy: Policy, policy_factory: PolicyFactory) -> Policy: ...
    @overload
    def minimize(self, policy: Policy, true_state_pairs: List[Tuple[State, State]], false_state_pairs: List[Tuple[State, State]], policy_factory: PolicyFactory) -> Policy: ...



from dlplan.core import VocabularyInfo, InstanceInfo, SyntacticElementFactory, State


def generate_bw_vocabulary():
    vocabulary = VocabularyInfo()
    vocabulary.add_predicate("on", 2)
    vocabulary.add_predicate("on_g", 2)
    vocabulary.add_predicate("ontable", 1)
    vocabulary.add_predicate("holding", 1)
    vocabulary.add_predicate("clear", 1)
    return vocabulary


def generate_bw_instance(vocabulary):
    instance = InstanceInfo(-1, vocabulary)
    instance.add_atom("on", ["a", "b"])
    instance.add_atom("on", ["b", "a"])
    instance.add_atom("ontable", ["a"])
    instance.add_atom("ontable", ["b"])
    instance.add_atom("holding", ["a"])
    instance.add_atom("holding", ["b"])
    instance.add_atom("clear", ["a"])
    instance.add_atom("clear", ["b"])
    instance.add_static_atom("on_g", ["a", "b"])
    return instance


def test_vocabulary():
    """ Exhaustively test VocabularyInfo
    """
    vocabulary = generate_bw_vocabulary()
    predicates = vocabulary.get_predicates()
    assert predicates[0].get_name() == "on"
    assert predicates[0].get_index() == 0
    assert predicates[0].get_arity() == 2
    assert predicates[1].get_name() == "on_g"
    assert predicates[1].get_index() == 1
    assert predicates[1].get_arity() == 2
    assert predicates[2].get_name() == "ontable"
    assert predicates[2].get_index() == 2
    assert predicates[2].get_arity() == 1
    assert predicates[3].get_name() == "holding"
    assert predicates[3].get_index() == 3
    assert predicates[3].get_arity() == 1
    assert predicates[4].get_name() == "clear"
    assert predicates[4].get_index() == 4
    assert predicates[4].get_arity() == 1


def test_instance():
    """ Exhaustively test InstanceInfo
    """
    vocabulary = generate_bw_vocabulary()
    instance = generate_bw_instance(vocabulary)

    objects = instance.get_objects()
    assert objects[0].get_name() == "a"
    assert objects[0].get_index() == 0
    assert objects[1].get_name() == "b"
    assert objects[1].get_index() == 1

    atoms = instance.get_atoms()
    assert atoms[0].get_name() == "on(a,b)"
    assert atoms[0].get_index() == 0
    assert [objects[obj].get_name() for obj in atoms[0].get_object_indices()] == ["a", "b"]
    assert atoms[1].get_name() == "on(b,a)"
    assert atoms[1].get_index() == 1
    assert [objects[obj].get_name() for obj in atoms[1].get_object_indices()] == ["b", "a"]
    assert atoms[2].get_name() == "ontable(a)"
    assert atoms[2].get_index() == 2
    assert [objects[obj].get_name() for obj in atoms[2].get_object_indices()] == ["a"]


def test_factory():
    """ Test the construction, evaluation, and getters of Elements.
    """
    vocabulary = generate_bw_vocabulary()
    instance = generate_bw_instance(vocabulary)
    factory = SyntacticElementFactory(vocabulary)

    atoms = instance.get_atoms()
    state = State(-1, instance, [atoms[0], atoms[3], atoms[6]])
    assert str(state) == "(instance index=-1, state index=-1, atoms={on(a,b), ontable(b), clear(a)})"

    numerical = factory.parse_numerical("n_count(c_and(c_primitive(on_g,0),c_primitive(on,0)))")
    assert numerical.compute_complexity() == 4
    assert numerical.evaluate(state) == 1

    boolean = factory.parse_boolean("b_empty(c_and(c_primitive(on_g,0),c_primitive(on,0)))")
    assert boolean.compute_complexity() == 4
    assert boolean.evaluate(state) == 0

    numerical = factory.parse_numerical("n_count(c_projection(r_primitive(on_g,0,1), 1))")
    assert numerical.compute_complexity() == 3
    assert numerical.evaluate(state) == 1



from dlplan.core import VocabularyInfo, InstanceInfo, \
    SyntacticElementFactory, State, DenotationsCaches

from ..utils.denotation import create_concept_denotation, create_role_denotation


def test_caching():
    vocabulary = VocabularyInfo()
    predicate_0 = vocabulary.add_predicate("role", 2)
    instance = InstanceInfo(0, vocabulary)
    atom_0 = instance.add_atom("role", ["A", "B"])

    state_0 = State(0, instance, [])
    state_1 = State(1, instance, [atom_0])

    factory = SyntacticElementFactory(vocabulary)
    caches = DenotationsCaches()

    concept_0 = factory.parse_concept("c_primitive(role, 0)")
    assert concept_0.evaluate(state_0) == create_concept_denotation(instance, [])
    assert concept_0.evaluate(state_0) == concept_0.evaluate(state_0, caches)
    assert concept_0.evaluate(state_0, caches) == concept_0.evaluate(state_0, caches)
    assert concept_0.evaluate(state_1) == create_concept_denotation(instance, ["A"])
    assert concept_0.evaluate(state_1) == concept_0.evaluate(state_1, caches)
    assert concept_0.evaluate(state_1, caches) == concept_0.evaluate(state_1, caches)
    assert [concept_0.evaluate(state_0,caches), concept_0.evaluate(state_1,caches)] == \
        concept_0.evaluate([state_0, state_1], caches)

    role_0 = factory.parse_role("r_primitive(role, 0, 1)")
    assert role_0.evaluate(state_0) == create_role_denotation(instance, [])
    assert role_0.evaluate(state_0) == role_0.evaluate(state_0, caches)
    assert role_0.evaluate(state_0, caches) == role_0.evaluate(state_0, caches)
    assert role_0.evaluate(state_1) == create_role_denotation(instance, [("A", "B")])
    assert role_0.evaluate(state_1) == role_0.evaluate(state_1, caches)
    assert role_0.evaluate(state_1, caches) == role_0.evaluate(state_1, caches)
    assert [role_0.evaluate(state_0,caches), role_0.evaluate(state_1,caches)] == \
        role_0.evaluate([state_0, state_1], caches)

    numerical_0 = factory.parse_numerical("n_count(c_primitive(role, 0))")
    assert numerical_0.evaluate(state_0) == 0
    assert numerical_0.evaluate(state_0) == numerical_0.evaluate(state_0, caches)
    assert numerical_0.evaluate(state_1) == 1
    assert numerical_0.evaluate(state_1) == numerical_0.evaluate(state_1, caches)
    assert [numerical_0.evaluate(state_0), numerical_0.evaluate(state_1)] == \
        numerical_0.evaluate([state_0, state_1], caches)
    assert numerical_0.evaluate([state_0, state_1], caches) == numerical_0.evaluate([state_0, state_1], caches)

    boolean_0 = factory.parse_boolean("b_empty(c_primitive(role, 0))")
    assert boolean_0.evaluate(state_0) is True
    assert boolean_0.evaluate(state_0) == boolean_0.evaluate(state_0, caches)
    assert boolean_0.evaluate(state_1) is False
    assert boolean_0.evaluate(state_1) == boolean_0.evaluate(state_1, caches)
    assert [boolean_0.evaluate(state_0), boolean_0.evaluate(state_1)] == \
        boolean_0.evaluate([state_0, state_1], caches)
    assert boolean_0.evaluate([state_0, state_1], caches) == boolean_0.evaluate([state_0, state_1], caches)



#!/usr/bin/python3

"""
Example illustrating the policy component.
"""

from dlplan.core import VocabularyInfo, InstanceInfo, State, SyntacticElementFactory, \
    DenotationsCaches
from dlplan.policy import PolicyFactory, PositiveBooleanEffect, DecrementNumericalEffect


def construct_vocabulary_info():
    """ Construct a VocabularyInfo for the Blocks domain.

    Create an empty VocabularyInfo and then add predicates, and constants.

    Returns: the VocabularyInfo
    """
    vocabulary = VocabularyInfo()
    vocabulary.add_predicate("on", 2)
    vocabulary.add_predicate("on_g", 2)
    vocabulary.add_predicate("ontable", 1)
    vocabulary.add_predicate("holding", 1)
    vocabulary.add_predicate("clear", 1)
    vocabulary.add_predicate("arm-empty", 0)
    return vocabulary


def construct_instance_info(vocabulary):
    """ Construct an InstanceInfo over the Blocks domain.

    Create an empty InstanceInfo and then add objects, atoms, and static atoms.

    Returns: the InstanceInfo
    """
    # User must ensure that each index gets its unique index for caching.
    instance = InstanceInfo(0, vocabulary)
    instance.add_atom("on", ["a", "b"])
    instance.add_atom("on", ["b", "a"])
    instance.add_atom("ontable", ["a"])
    instance.add_atom("ontable", ["b"])
    instance.add_atom("holding", ["a"])
    instance.add_atom("holding", ["b"])
    instance.add_atom("clear", ["a"])
    instance.add_atom("clear", ["b"])
    instance.add_atom("arm-empty", [])
    instance.add_static_atom("on_g", ["a", "b"])
    return instance


def main():
    vocabulary = construct_vocabulary_info()
    instance = construct_instance_info(vocabulary)
    element_factory = SyntacticElementFactory(vocabulary)

    print("Constructing policy:")
    policy_factory = PolicyFactory(element_factory)
    # boolean represents whether the hand is empty or not
    boolean = policy_factory.make_boolean("b0", element_factory.parse_boolean("b_nullary(arm-empty)"))
    # numerical representes the number of blocks on top of another block
    numerical = policy_factory.make_numerical("n1", element_factory.parse_numerical("n_count(r_primitive(on,0,1))"))
    b_pos_condition_0 = policy_factory.make_pos_condition(boolean)
    b_bot_effect_0 = policy_factory.make_bot_effect(boolean)
    n_gt_condition_0 = policy_factory.make_gt_condition(numerical)
    n_dec_effect_0 = policy_factory.make_dec_effect(numerical)
    rule_1 = policy_factory.make_rule(
        {b_pos_condition_0, n_gt_condition_0},
        {b_bot_effect_0, n_dec_effect_0}
    )
    policy = policy_factory.make_policy({rule_1})

    atoms = instance.get_atoms()
    atom_0 = atoms[0]
    atom_1 = atoms[1]
    atom_2 = atoms[2]
    atom_3 = atoms[3]
    atom_6 = atoms[6]
    atom_7 = atoms[7]
    atom_8 = atoms[8]
    # User must ensure that each State gets its unique index for caching.
    state_0 = State(0, instance, [atom_0, atom_3, atom_6, atom_8])  # a on b
    state_1 = State(1, instance, [atom_1, atom_2, atom_7, atom_8])  # b on a
    state_2 = State(2, instance, [atom_2, atom_3, atom_6, atom_7, atom_8])  # a,b on table

    caches = DenotationsCaches()
    assert policy.evaluate(state_0, state_2, caches)
    assert policy.evaluate(state_1, state_2, caches)
    assert not policy.evaluate(state_2, state_0, caches)
    assert not policy.evaluate(state_2, state_1, caches)

    print(str(policy))
    print()

    print("Parsing policy:")
    policy_str = "(:policy\n" \
        "(:booleans (b0 \"b_nullary(arm-empty)\"))\n" \
        "(:numericals (n0 \"n_count(r_primitive(on,0,1))\"))\n" \
        "(:rule (:conditions (:c_b_pos b0) (:c_n_gt n0)) (:effects (:e_b_bot b0) (:e_n_dec n0)))\n" \
        ")"
    policy_in = policy_factory.parse_policy(policy_str)
    print(str(policy_in))
    print()


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
""" Module description: library initialization using the PDDL parsing
    functionality of Tarski (https://tarski.readthedocs.io/en/latest/)
"""

import argparse
import time
import sys

from domain_data import DomainData
from instance_data import InstanceData


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Generating features for a single PDDL instance.')
    parser.add_argument("--domain", type=str, required=True)
    parser.add_argument("--instance", type=str, required=True)
    parser.add_argument("--time_limit", type=int, required=False, default=float("inf"))
    parser.add_argument("--c", type=int, required=False, default=5)
    parser.add_argument("--t", type=int, required=False, default=1800)
    parser.add_argument("--f", type=int, required=False, default=1000000)
    parser.add_argument("--n", type=int, required=False, default=1)
    args = parser.parse_args()

    domain_data = DomainData(args.domain)
    instance_data = InstanceData(args.instance, domain_data, args.time_limit)

    # print vocabulary and instance information for easy re-parsing from the cpp code.
    print("Predicates:")
    print("\n".join([str(predicate) for predicate in domain_data.vocabulary_info.get_predicates()]))
    print("\nConstants:")
    print("\n".join([str(constant) for constant in domain_data.vocabulary_info.get_constants()]))
    print("\nObjects:")
    print("\n".join([str(obj) for obj in instance_data.instance_info.get_objects()]))
    print("\nAtoms:")
    print("\n".join([str(obj) for obj in instance_data.instance_info.get_atoms()]))
    print("\nStatic atom indices:")
    print("\n".join([str(idx) for idx in instance_data.instance_info.get_static_atom_idxs()]))
    print("\nStates:")
    print("\n".join([str(state) for state in instance_data.dlplan_states]))
    print()

    print("Instance information:")
    print(f"Number of states: {len(instance_data.dlplan_states)}")

    sys.stdout.flush()

    feature_reprs = domain_data.feature_generator.generate(domain_data.syntactic_element_factory, args.c, args.t, args.f, args.n, instance_data.dlplan_states)
    start = time.time()
    for feature_repr in feature_reprs:
        if feature_repr.startswith("n_"):
            f = domain_data.syntactic_element_factory.parse_numerical(feature_repr)
        elif feature_repr.startswith("b_"):
            f = domain_data.syntactic_element_factory.parse_boolean(feature_repr)
        elif feature_repr.startswith("c_"):
            f = domain_data.syntactic_element_factory.parse_concept(feature_repr)
        elif feature_repr.startswith("r_"):
            f = domain_data.syntactic_element_factory.parse_role(feature_repr)
        for dlplan_state in instance_data.dlplan_states:
            f.evaluate(dlplan_state)
    end = time.time()
    print(f"Time to construct and evaluate features: {end - start}s")



from dlplan.core import InstanceInfo, ConceptDenotation, RoleDenotation

from typing import List, Tuple


def create_concept_denotation(instance: InstanceInfo, object_names: List[str]):
    num_objects = len(instance.get_objects())
    denotation = ConceptDenotation(num_objects)
    for object_name in object_names:
        object = instance.get_object(object_name)
        denotation.insert(object.get_index())
    return denotation


def create_role_denotation(instance: InstanceInfo, pairs_of_object_names: List[Tuple[str, str]]):
    num_objects = len(instance.get_objects())
    denotation = RoleDenotation(num_objects)
    for object_name_1, object_name_2 in pairs_of_object_names:
        object_first = instance.get_object(object_name_1)
        object_second = instance.get_object(object_name_2)
        denotation.insert((object_first.get_index(), object_second.get_index()))
    return denotation



#!/usr/bin/python3

"""
Example illustrating the state_space component.
"""

from dlplan.state_space import generate_state_space


def main():
    """ Generate state spaces over the same vocabulary and compute distances.
    """
    # Generate and read state space files.
    state_space_2_1_0 = generate_state_space("domain.pddl", "instance_2_1_0.pddl").state_space
    vocabulary = state_space_2_1_0.get_instance_info().get_vocabulary_info()
    print("State space of instance_2_1_0:")
    print(str(state_space_2_1_0))

    # Generate and read state space files over same VocabularyInfo.
    state_space_2_2_0 = generate_state_space("domain.pddl", "instance_2_1_0.pddl", vocabulary).state_space
    print("State space of instance_2_2_0:")
    print(str(state_space_2_2_0))
    print()

    # Compute goal distances and deadends
    goal_distances = state_space_2_1_0.compute_goal_distances()
    print("Goal distances of instance_2_1_0:")
    for state_index, distance in goal_distances.items():
        print("state_index=", state_index, "distance=", distance)
    print("Deadends:")
    for state_index in state_space_2_1_0.get_states().keys():
        if state_index not in goal_distances:
            print(state_index, end=" ")
    print()

    # Compute forward distances from states with options forward=true, stop_if_goal=false
    forward_distances = state_space_2_1_0.compute_distances(
        set([state_space_2_1_0.get_initial_state_index()]),
        True,
        False)
    print("Forward distances from initial state of instance_2_1_0:")
    for state_index, distance in forward_distances.items():
        print("state_index=", state_index, "distance=", distance)


if __name__ == "__main__":
    main()



from pathlib import Path

from dlplan.state_space import generate_state_space, GeneratorExitCode


HERE = Path(__file__).resolve().parent


def test_state_space_gripper_incomplete():
    """ """
    result = generate_state_space(str(HERE / "domain.pddl"), str(HERE / "p-1-0.pddl"), None, -1, 10)
    assert result.exit_code == GeneratorExitCode.COMPLETE



from _dlplan import NamedBoolean, NamedNumerical, NamedConcept, NamedRole, \
    BaseCondition, BaseEffect, Rule, Policy, PolicyFactory, \
    PolicyMinimizer, \
    PositiveBooleanCondition, NegativeBooleanCondition, \
    GreaterNumericalCondition, EqualNumericalCondition, \
    PositiveBooleanEffect, NegativeBooleanEffect, UnchangedBooleanEffect, \
    DecrementNumericalEffect, IncrementNumericalEffect, UnchangedNumericalEffect



from typing import List

from ..core import SyntacticElementFactory, State


class FeatureGenerator:
    def generate(self, 
        factory: SyntacticElementFactory, 
        states: List[State],
        concept_complexity_limit: int = 9,
        role_complexity_limit: int = 9,
        boolean_complexity_limit: int = 9,
        count_numerical_complexity_limit: int = 9,
        distance_numerical_complexity_limit: int = 9,
        time_limit: int = 3600,
        feature_limit: int = 10000) -> List[str]: ...
    def set_generate_empty_boolean(self, enable: bool) -> None: ...
    def set_generate_inclusion_boolean(self, enable: bool) -> None: ...
    def set_generate_nullary_boolean(self, enable: bool) -> None: ...
    def set_generate_all_concept(self, enable: bool) -> None: ...
    def set_generate_and_concept(self, enable: bool) -> None: ...
    def set_generate_bot_concept(self, enable: bool) -> None: ...
    def set_generate_diff_concept(self, enable: bool) -> None: ...
    def set_generate_equal_concept(self, enable: bool) -> None: ...
    def set_generate_not_concept(self, enable: bool) -> None: ...
    def set_generate_one_of_concept(self, enable: bool) -> None: ...
    def set_generate_or_concept(self, enable: bool) -> None: ...
    def set_generate_primitive_concept(self, enable: bool) -> None: ...
    def set_generate_projection_concept(self, enable: bool) -> None: ...
    def set_generate_some_concept(self, enable: bool) -> None: ...
    def set_generate_subset_concept(self, enable: bool) -> None: ...
    def set_generate_top_concept(self, enable: bool) -> None: ...
    def set_generate_concept_distance_numerical(self, enable: bool) -> None: ...
    def set_generate_count_numerical(self, enable: bool) -> None: ...
    def set_generate_and_role(self, enable: bool) -> None: ...
    def set_generate_compose_role(self, enable: bool) -> None: ...
    def set_generate_diff_role(self, enable: bool) -> None: ...
    def set_generate_identity_role(self, enable: bool) -> None: ...
    def set_generate_inverse_role(self, enable: bool) -> None: ...
    def set_generate_not_role(self, enable: bool) -> None: ...
    def set_generate_or_role(self, enable: bool) -> None: ...
    def set_generate_primitive_role(self, enable: bool) -> None: ...
    def set_generate_restrict_role(self, enable: bool) -> None: ...
    def set_generate_top_role(self, enable: bool) -> None: ...
    def set_generate_transitive_closure_role(self, enable: bool) -> None: ...
    def set_generate_transitive_reflexive_closure_role(self, enable: bool) -> None: ...


def generate_features(self, 
    factory: SyntacticElementFactory,
    states: List[State],
    concept_complexity_limit: int = 9,
    role_complexity_limit: int = 9,
    boolean_complexity_limit: int = 9,
    count_numerical_complexity_limit: int = 9,
    distance_numerical_complexity_limit: int = 9,
    time_limit: int = 3600,
    feature_limit: int = 10000,
    generate_empty_boolean: bool = True,
    generate_inclusion_boolean: bool = True,
    generate_nullary_boolean: bool = True,
    generate_all_concept: bool = True,
    generate_and_concept: bool = True,
    generate_bot_concept: bool = True,
    generate_diff_concept: bool = False,
    generate_equal_concept: bool = True,
    generate_not_concept: bool = True,
    generate_one_of_concept: bool = True,
    generate_or_concept: bool = False,
    generate_primitive_concept: bool = True,
    generate_projection_concept: bool = False,
    generate_some_concept: bool = True,
    generate_subset_concept: bool = False,
    generate_top_concept: bool = True,
    generate_concept_distance_numerical: bool = True,
    generate_count_numerical: bool = True,
    generate_and_role: bool = True,
    generate_compose_role: bool = False,
    generate_diff_role: bool = False,
    generate_identity_role: bool = True,
    generate_inverse_role: bool = True,
    generate_not_role: bool = False,
    generate_or_role: bool = False,
    generate_primitive_role: bool = True,
    generate_restrict_role: bool = True,
    generate_top_role: bool = False,
    generate_transitive_closure_role: bool = True,
    generate_transitive_reflexive_closure_role: bool = False) -> List[str]: ...



from dlplan.core import VocabularyInfo, InstanceInfo, SyntacticElementFactory, State
from dlplan.generator import FeatureGenerator


def generate_bw_vocabulary():
    vocabulary = VocabularyInfo()
    vocabulary.add_predicate("on", 2)
    vocabulary.add_predicate("on_g", 2)
    vocabulary.add_predicate("ontable", 1)
    vocabulary.add_predicate("holding", 1)
    vocabulary.add_predicate("clear", 1)
    return vocabulary


def generate_bw_instance(vocabulary):
    instance = InstanceInfo(0, vocabulary)
    instance.add_atom("on", ["a", "b"])
    instance.add_atom("on", ["b", "a"])
    instance.add_atom("ontable", ["a"])
    instance.add_atom("ontable", ["b"])
    instance.add_atom("holding", ["a"])
    instance.add_atom("holding", ["b"])
    instance.add_atom("clear", ["a"])
    instance.add_atom("clear", ["b"])
    instance.add_static_atom("on_g", ["a", "b"])
    return instance


def test_generate_exhaustively():
    vocabulary = generate_bw_vocabulary()
    instance = generate_bw_instance(vocabulary)
    factory = SyntacticElementFactory(vocabulary)
    generator = FeatureGenerator()

    atoms = instance.get_atoms()
    state_0 = State(0, instance, [atoms[0], atoms[3]])  # a on b
    state_1 = State(1, instance, [atoms[1], atoms[2]])  # b on a
    state_2 = State(2, instance, [atoms[2], atoms[3]])  # both on table
    states = [state_0, state_1, state_2]
    feature_reprs = generator.generate(factory, states, 2, 2, 5, 5, 5, 180, 100000)

    boolean_reprs = [repr for repr in feature_reprs if repr.startswith("b_")]
    numerical_reprs = [repr for repr in feature_reprs if repr.startswith("n_")]




from _dlplan import ConceptDenotation, RoleDenotation, DenotationsCaches, \
    Constant, Predicate, VocabularyInfo, Object, Atom, \
    InstanceInfo, State, Concept, Role, Boolean, Numerical, \
    SyntacticElementFactory




import os
import sys
import subprocess
import multiprocessing

from pathlib import Path

from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext

__version__ = "0.3.14"
HERE = Path(__file__).resolve().parent


# A CMakeExtension needs a sourcedir instead of a file list.
# The name must be the _single_ output extension from the CMake build.
# If you need multiple extensions, see scikit-build.
class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))

        # required for auto-detection of auxiliary "native" libs
        if not extdir.endswith(os.path.sep):
            extdir += os.path.sep

        cfg = "Debug" if self.debug else "Release"

        temp_directory = Path.cwd() / self.build_temp
        # Create the temporary build directory, if it does not already exist
        os.makedirs(temp_directory, exist_ok=True)

        # Build dependencies
        subprocess.run(
            ["cmake", "-S", f"{ext.sourcedir}/dependencies", "-B", f"{str(temp_directory)}/dependencies/build", f"-DCMAKE_INSTALL_PREFIX={str(temp_directory)}/dependencies/installs"], cwd=str(temp_directory), check=True
        )

        subprocess.run(
            ["cmake", "--build", f"{str(temp_directory)}/dependencies/build", "-j16"]
        )

        # Build dlplan
        cmake_args = [
            "-DDLPLAN_PYTHON=On",
            f"-DDLPLAN_VERSION_INFO={__version__}",
            f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY={extdir}",
            "-DBUILD_TESTS:bool=false",
            f"-DPYTHON_EXECUTABLE={sys.executable}",
            f"-DCMAKE_BUILD_TYPE={cfg}",  # not used on MSVC, but no harm
            f"-DCMAKE_PREFIX_PATH={str(temp_directory)}/dependencies/installs"
        ]
        build_args = []
        build_args += ["--target", ext.name]

        subprocess.run(
            ["cmake", "-S", ext.sourcedir, "-B", f"{str(temp_directory)}/build"] + cmake_args, cwd=str(temp_directory), check=True
        )
        subprocess.run(
            ["cmake", "--build", f"{str(temp_directory)}/build", f"-j{multiprocessing.cpu_count()}"] + build_args, cwd=str(temp_directory), check=True
        )


# The information here can also be placed in setup.cfg - better separation of
# logic and declaration, and simpler if you include description/version in a file.
setup(
    name="dlplan",
    version=__version__,
    author="Dominik Drexler, Jendrik Seipp and Guillem Francès",
    author_email="dominik.drexler@liu.se",
    url="https://github.com/rleap-project/dlplan",
    description="A library for using description logics features in planning",
    long_description="",
    install_requires=["state_space_generator==0.1.9", "cmake>=3.21"],
    packages=find_packages(where="api/python/src"),
    package_dir={"": "api/python/src"},
    package_data={
        "": ["*.pyi"],
    },
    ext_modules=[CMakeExtension("_dlplan")],
    cmdclass={"build_ext": CMakeBuild},
    zip_safe=False,
    extras_require={
        'test': [
            'pytest',
        ],
    }
)



from enum import Enum

from typing import Overload, Dict, List, MutableSet

from ..core import VocabularyInfo, InstanceInfo, State


class StateSpace:
    @overload
    def __init__(self,
        instance_info: InstanceInfo,
        states: Dict[int, State],
        initial_state_index: int,
        forward_successor_state_indices: Dict[int, MutableSet[int]],
        goal_state_indices: MutableSet[int]) -> None: ...
    @overload
    def __init__(self, state_space: "StateSpace", state_indices: MutableSet[int]) -> None: ...
    def __str__(self) -> str: ...
    def compute_distances(self, state_indices: MutableSet[int], forward: bool, stop_if_goal: bool) -> Dict[int, int]: ...
    def compute_goal_distances(self) -> Dict[int, int]: ...
    def is_goal(self, state_index: int) -> bool: ...
    def to_dot(self, verbosity_level: int) -> str: ...
    def set_initial_state_index(self, index: int) -> None: ...
    def set_goal_state_indices(self, state_indices: MutableSet[int]) -> None: ...
    def get_states(self) -> Dict[int, State]: ...
    def get_initial_state_index(self) -> int: ...
    def get_forward_successor_state_indices(self) -> Dict[int, MutableSet[int]]: ...
    def get_backward_successor_state_indices(self) -> Dict[int, MutableSet[int]]: ...
    def get_goal_state_indices(self) -> MutableSet[int]: ...
    def get_instance_info(self) -> InstanceInfo: ...


class GeneratorExitCode(Enum):
    COMPLETE = ...
    INCOMPLETE = ...
    FAIL = ...


class GeneratorResult:
    exit_code: GeneratorExitCode
    state_space: StateSpace


def generate_state_space(domain_file: str, instance_file: str, vocabulary_info: VocabularyInfo = None, index: int = -1, max_time: int = 2147483646, max_num_states: int = 2147483646) -> GeneratorResult: ...



#! /usr/bin/env python

"""
[t=0.152091s, 48508 KB] Complexity 5:
Total generated features: 399
Total novel features: 63
Total concept elements: 28
Total role elements: 5
Total numerical elements: 22
Total boolean elements: 8
"""

import re

from lab.parser import Parser


parser = Parser()
parser.add_pattern("num_iterations", r"Number of iterations: (.+)", type=int)
parser.add_pattern("num_states", r"Number of states: (.+)", type=int)
parser.add_pattern("num_dynamic_atoms", r"Number of dynamic atoms: (.+)", type=int)
parser.add_pattern("num_static_atoms", r"Number of static atoms: (.+)", type=int)
parser.add_pattern("generate_time_complexity_5", r"\[t=(.+)s, .+ KB\] Complexity 5:\n", type=float)
parser.add_pattern("generate_memory_complexity_5", r"\[t=.+s, (.+) KB\] Complexity 5:\n", type=int)
parser.add_pattern("generate_time_complexity_10", r"\[t=(.+)s, .+ KB\] Complexity 10:\n", type=float)
parser.add_pattern("generate_memory_complexity_10", r"\[t=.+s, (.+) KB\] Complexity 10:\n", type=int)
parser.add_pattern("num_booleans", r"Number of booleans: (.+)", type=int)
parser.add_pattern("num_numericals", r"Number of numericals: (.+)", type=int)
parser.add_pattern("evaluate_time_single", r"Time evaluate features for single state without cache: (.+)ms", type=float)
parser.add_pattern("evaluate_time_single_cache", r"Time evaluate features for single state with cache: (.+)ms", type=float)
parser.add_pattern("evaluate_time_multi_cache", r"Time evaluate features for all states with cache: (.+)ms", type=float)
parser.parse()
